import dataclasses

import numpy as np

from . import Model


@dataclasses.dataclass(kw_only=True)
class Hoerl(Model):
    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        return a * b**t * t**c

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        return a * b**t * t ** (c - 1) * (np.log(b) * t + c)



import typing

import pandas as pd
from rich import print
from rich.console import RenderableType
from rich.padding import PaddingDimensions
from rich.panel import Panel
from rich.pretty import Pretty
from rich.pretty import install as pretty_install
from rich.style import Style
from rich.table import Table
from rich.text import Text
from rich.traceback import install as traceback_install

pretty_install()
traceback_install()
panel_title_style = Style(color="bright_blue", bold=True)


def make_panel(
    obj: RenderableType,
    *args,
    title: typing.Optional[str] = None,
    expand: bool = False,
    **kwargs,
) -> Panel:
    panel = Panel(
        obj,
        *args,
        title=Text(text=title, style=panel_title_style) if title else None,
        expand=expand,
        **kwargs,
    )
    return panel


def print_panel(
    obj: RenderableType,
    *args,
    title: typing.Optional[str] = None,
    expand: bool = False,
    **kwargs,
) -> None:
    panel = make_panel(obj=obj, *args, title=title, expand=expand, **kwargs)
    print(panel)


def make_data_frame(
    data_frame: pd.DataFrame,
    *args,
    title: typing.Optional[str] = None,
    expand: bool = False,
    padding: PaddingDimensions = (0, 4),
    **kwargs,
) -> Panel:
    table = Table.grid(padding=padding)
    for key, values in data_frame.items():
        table.add_row(str(key), *map(Pretty, values))
    panel = make_panel(table, *args, title=title, expand=expand, **kwargs)
    return panel


def print_data_frame(
    data_frame: pd.DataFrame,
    *args,
    title: typing.Optional[str] = None,
    expand: bool = False,
    padding: PaddingDimensions = (0, 4),
    **kwargs,
) -> None:
    panel = make_data_frame(
        data_frame=data_frame,
        *args,
        title=title,
        expand=expand,
        padding=padding,
        **kwargs,
    )
    print(panel)


def make_series(
    series: pd.Series,
    *args,
    title: typing.Optional[str] = None,
    expand: bool = False,
    padding: PaddingDimensions = (0, 4),
    **kwargs,
) -> Panel:
    table = Table.grid(padding=padding)
    for key, value in series.items():
        table.add_row(str(key), Pretty(value))
    panel = make_panel(table, *args, title=title, expand=expand, **kwargs)
    return panel


def print_series(
    series: pd.Series,
    *args,
    title: typing.Optional[str] = None,
    expand: bool = False,
    padding: PaddingDimensions = (0, 4),
    **kwargs,
) -> None:
    panel = make_series(
        series=series, *args, title=title, expand=expand, padding=padding, **kwargs
    )
    print(panel)


def make_dict(
    obj: dict,
    *args,
    title: typing.Optional[str] = None,
    expand: bool = False,
    padding: PaddingDimensions = (0, 4),
    **kwargs,
) -> Panel:
    series = pd.Series(obj)
    panel = make_series(
        series=series, *args, title=title, expand=expand, padding=padding, **kwargs
    )
    return panel


def print_dict(
    obj: dict,
    *args,
    title: typing.Optional[str] = None,
    expand: bool = False,
    padding: PaddingDimensions = (0, 4),
    **kwargs,
) -> None:
    panel = make_dict(
        obj=obj, *args, title=title, expand=expand, padding=padding, **kwargs
    )
    print(panel)



import dataclasses

import numpy as np
import numpy.typing as npt
import scipy.stats

from . import Model


@dataclasses.dataclass(kw_only=True)
class InverseGaussian(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=([0, 0], np.inf)
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        return scipy.stats.invgauss.cdf(x=t, mu=b / a, scale=a)

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        return scipy.stats.invgauss.pdf(x=t, mu=b / a, scale=a)



import dataclasses

import numpy as np
import numpy.typing as npt
import scipy.special

from . import Model


@dataclasses.dataclass(kw_only=True)
class ShiftedLogPearson3(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=([0, 1, 0], np.inf)
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        s = scipy.special.gammainc(b, c * np.log(t / a + 1))
        return s

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        u = (
            c**b
            * (np.log(t / a + 1)) ** (b - 1)
            / (a * scipy.special.gamma(b) * (t / a + 1) ** (c + 1))
        )
        return u



import dataclasses
import typing

import numpy as np
import numpy.typing as npt

FUNCTIONS = ["cdf", "pdf"]


@dataclasses.dataclass(kw_only=True)
class Model:
    p0: typing.Optional[npt.ArrayLike] = None
    sigma: typing.Optional[npt.ArrayLike] = None
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=(-np.inf, np.inf)
    )

    @classmethod
    @property
    def name(cls) -> str:
        return cls.__name__

    @staticmethod
    def cdf(t: float | np.ndarray, *popt: float) -> float | np.ndarray:
        raise NotImplementedError()

    @staticmethod
    def pdf(t: float | np.ndarray, *popt: float) -> float | np.ndarray:
        raise NotImplementedError()

    @classmethod
    def forward(
        cls, t: float | np.ndarray, *popt: float, func: str = "cdf"
    ) -> float | np.ndarray:
        if func == "cdf":
            return cls.cdf(t, *popt)
        elif func == "pdf":
            return cls.pdf(t, *popt)
        else:
            raise NotImplementedError()

    @classmethod
    def select(cls, func: str = "cdf"):
        if func == "cdf":
            return cls.cdf
        elif func == "pdf":
            return cls.pdf
        else:
            raise NotImplementedError()

    def prepare(
        self, x: np.ndarray, y: np.ndarray, func: str = "cdf"
    ) -> tuple[np.ndarray, np.ndarray]:
        return x, y



from .. import Model
from .beta import Beta
from .double_power import DoublePower
from .double_triangular import DoubleTriangular
from .frechet import Frechet
from .gamma import Gamma
from .hill import Hill
from .hoerl import Hoerl
from .inverse_gaussian import InverseGaussian
from .kumaraswamy import Kumaraswamy
from .logistic import Logistic
from .multistage import Multistage
from .normal_gaussian import NormalGaussian
from .polynomial import Polynomial
from .rational import Rational
from .shiftred_log_pearson_3 import ShiftedLogPearson3
from .weibull import Weibull

NAME_TO_MODEL = {
    "Beta": Beta,
    "DoublePower": DoublePower,
    "DoubleTriangular": DoubleTriangular,
    "Frechet": Frechet,
    "Gamma": Gamma,
    "Hill": Hill,
    "Hoerl": Hoerl,
    "InverseGaussian": InverseGaussian,
    "Kumaraswamy": Kumaraswamy,
    "Logistic": Logistic,
    "Multistage": Multistage,
    "NormalGaussian": NormalGaussian,
    "Polynomial": Polynomial,
    "Rational": Rational,
    "ShiftedLogPearson3": ShiftedLogPearson3,
    "Weibull": Weibull,
}



import dataclasses

import numpy as np

from . import Model


@dataclasses.dataclass(kw_only=True)
class Hill(Model):
    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        s = a * t**b / (c**b + t**b)
        return s

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        u = a * b * c**b * t ** (b - 1) / (c**b + t**b) ** 2
        return u



import glob
import os
import typing

import click
import numpy as np
import pandas as pd
from rich.columns import Columns

from ..log import make_dict, make_series, print
from . import plot_params_distribution


@click.command(name="param-dis")
@click.option("--input-dir", type=click.Path(), default="results/params/cdf")
@click.option("--output-dir", type=click.Path(), default="results/param-dis/cdf/cdf")
@click.option("--key", default="cdf_r2_score")
@click.option(
    "--threshold-percentage", type=click.FloatRange(min=0, max=100), default=None
)
@click.option(
    "--threshold", show_default=True, type=click.FloatRange(min=0, max=1), default=0.9
)
def main(
    input_dir: str = "results/params/cdf",
    output_dir: str = "results/param-dis/cdf/cdf",
    key: str = "cdf_r2_score",
    threshold_percentage: typing.Optional[float] = None,
    threshold: float = 0.9,
):
    files = glob.glob(pathname=f"{input_dir}/*.csv")
    dfs: dict[str, pd.DataFrame] = dict()
    for filepath in files:
        name, _ = os.path.splitext(os.path.basename(filepath))
        df = pd.read_csv(filepath)
        df["model"] = name
        dfs[name] = df
    r2_score = pd.concat(
        [df[["id", "model", key]] for df in dfs.values()], ignore_index=True
    )
    columns = Columns()
    columns.add_renderable(
        make_series(r2_score[key].describe(), title="R2 Score Describe (All)")
    )
    if threshold_percentage:
        threshold = float(np.percentile(a=r2_score[key], q=threshold_percentage))
        columns.add_renderable(
            make_dict(
                {"percentage": threshold_percentage, key: threshold},
                title="R2 Score Threshold",
            )
        )
    else:
        columns.add_renderable(
            make_dict(
                {key: threshold},
                title="R2 Score Threshold",
            )
        )
    r2_score = r2_score[r2_score[key] > threshold]
    columns.add_renderable(
        make_series(r2_score[key].describe(), title="R2 Score Describe (Selected)")
    )
    columns.add_renderable(
        make_series(
            r2_score["model"].value_counts(), title="Catchment Counts (Selected)"
        )
    )
    print(columns)
    for name, df in dfs.items():
        df = df[df[key] > threshold]
        if df.empty:
            continue
        os.makedirs(name=output_dir, exist_ok=True)
        plot_params_distribution(data=df, fname=os.path.join(output_dir, f"{name}.png"))


if __name__ == "__main__":
    main()



import concurrent.futures
import glob
import os
import signal
import sys
from pathlib import Path

import click
import pandas as pd
from rich.columns import Columns
from rich.console import Console, Group
from rich.live import Live
from rich.panel import Panel
from rich.pretty import Pretty
from rich.progress import (
    BarColumn,
    MofNCompleteColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
)

from ..log import make_series
from .fitting import curve_fitting_file
from .model import FUNCTIONS, Model
from .models import NAME_TO_MODEL


def initializer():
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    sys.stderr = open(file=os.devnull, mode="w")


def save_results(
    results: list[dict],
    filepath: str,
    model_name: str,
    fit: str = "cdf",
    preds: list[str] = FUNCTIONS,
    console: Console = Console(),
) -> None:
    if not results:
        return
    df = pd.DataFrame.from_records(data=results, index="id")
    df.sort_index(inplace=True)
    columns = Columns()
    for pred in preds:
        panel = make_series(
            series=df[pred + "_r2_score"].describe(),
            title=f"{model_name} {fit} -> {pred} R2 Score",
        )
        columns.add_renderable(panel)
    console.print(columns)
    os.makedirs(name=os.path.dirname(filepath), exist_ok=True)
    df.to_csv(filepath)


def main(
    input_dir: str | Path = "2-sub-WFIUH_rescaled",
    output_dir: str | Path = "results/params/cdf",
    models: list[Model] = list(NAME_TO_MODEL.values()),
    fit: str = "cdf",
    preds: list[str] = FUNCTIONS,
    verbose: bool = False,
) -> int:
    files = glob.glob(pathname=f"{input_dir}/*.csv")
    overall_progress = Progress(
        TextColumn(text_format="{task.description}", style="logging.level.info"),
        BarColumn(),
        TaskProgressColumn(),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
    )
    models_progress = Progress(
        TextColumn(text_format="{task.description}", style="logging.level.info"),
        BarColumn(),
        TaskProgressColumn(),
        MofNCompleteColumn(),
        TimeElapsedColumn(),
        TimeRemainingColumn(),
    )
    progress_group = Group(
        Panel(models_progress, expand=False),
        Panel(overall_progress, expand=False),
    )
    with Live(progress_group) as live:
        overview_task_id = overall_progress.add_task(description="Overall Progress")
        for model in overall_progress.track(models, task_id=overview_task_id):
            model_task_id = models_progress.add_task(
                description=f"{model.name} {fit}", total=len(files)
            )
            results: list[dict] = []
            try:
                with concurrent.futures.ProcessPoolExecutor(
                    initializer=initializer
                ) as pool:
                    futures: list[concurrent.futures.Future] = list(
                        map(
                            lambda filepath: pool.submit(
                                curve_fitting_file,
                                model=model,
                                filepath=filepath,
                                fit=fit,
                                preds=preds.copy(),
                            ),
                            files,
                        )
                    )
                    for future in concurrent.futures.as_completed(futures):
                        try:
                            result = future.result()
                            if result:
                                results.append(result)
                                models_progress.advance(task_id=model_task_id)
                        except KeyboardInterrupt as e:
                            raise e
                        except Exception as e:
                            if verbose:
                                live.console.log(
                                    model.name,
                                    fit,
                                    Pretty(e),
                                    style="logging.level.error",
                                )
            except KeyboardInterrupt as e:
                save_results(
                    results,
                    os.path.join(output_dir, f"{model.name}.csv"),
                    model_name=model.name,
                    fit=fit,
                    preds=preds,
                    console=live.console,
                )
                raise e
            except Exception as e:
                live.console.print_exception()
            else:
                save_results(
                    results,
                    os.path.join(output_dir, f"{model.name}.csv"),
                    model_name=model.name,
                    fit=fit,
                    preds=preds,
                    console=live.console,
                )
            finally:
                models_progress.stop_task(model_task_id)
    return 0


@click.command(name="curve-fitting")
@click.option("--input-dir", type=click.Path(), default="2-sub-WFIUH_rescaled")
@click.option("--output-dir", type=click.Path(), default="results/params/cdf")
@click.option(
    "--model",
    type=click.Choice(choices=list(NAME_TO_MODEL.keys())),
    default=list(NAME_TO_MODEL.keys()),
    multiple=True,
)
@click.option("--fit", type=click.Choice(choices=FUNCTIONS), default="cdf")
@click.option(
    "--pred", type=click.Choice(choices=FUNCTIONS), default=FUNCTIONS, multiple=True
)
@click.option("--verbose", is_flag=True)
def command(
    input_dir: str | Path = "2-sub-WFIUH_rescaled",
    output_dir: str | Path = "results/params/cdf",
    model: tuple[str] = tuple(NAME_TO_MODEL.keys()),
    fit: str = "cdf",
    pred: tuple[str] = tuple(FUNCTIONS),
    verbose: bool = False,
) -> None:
    main(
        input_dir=input_dir,
        output_dir=output_dir,
        models=[NAME_TO_MODEL[name]() for name in model],
        fit=fit,
        preds=list(pred),
        verbose=verbose,
    )


if __name__ == "__main__":
    command()



import itertools
from pathlib import Path

import numpy as np
import scipy.optimize
import sklearn.metrics

from .model import FUNCTIONS, Model
from .utils import prepare_y, read_csv


def curve_fitting(
    model: Model,
    x: np.ndarray,
    y: np.ndarray,
    fits: list[str] = ["cdf"],
    preds: list[str] = FUNCTIONS,
) -> dict:
    popts: dict = dict()
    for fit in fits:
        x, y = model.prepare(x, y, func=fit)
        popt, pcov = scipy.optimize.curve_fit(
            f=model.select(fit),
            xdata=x,
            ydata=prepare_y(x=x, y=y, func=fit),
            p0=model.p0,
            sigma=model.sigma,
            bounds=model.bounds,
        )
        popts[fit] = popt
    r2_scores: dict = dict()
    for fit, pred in itertools.product(fits, preds):
        try:
            r2_score = sklearn.metrics.r2_score(
                y_true=prepare_y(x=x, y=y, func=pred),
                y_pred=model.select(func=pred)(x, *popts[fit]),
            )
            r2_scores[fit, pred] = r2_score
        except:
            pass
    return {"popts": popts, "r2_scores": r2_scores}


def curve_fitting_file(
    model: Model,
    filepath: str | Path,
    fit: str = "cdf",
    preds: list[str] = FUNCTIONS,
) -> dict:
    id, x, y = read_csv(filepath=filepath)
    res = curve_fitting(model=model, x=x, y=y, fits=[fit], preds=preds)
    return {
        "id": id,
        "fit": fit,
        **{chr(ord("a") + i): p for i, p in enumerate(res["popts"][fit])},
        **{
            pred + "_r2_score": r2_score
            for (fit, pred), r2_score in res["r2_scores"].items()
        },
    }



import dataclasses

import numpy as np
import numpy.typing as npt

from . import Model


@dataclasses.dataclass(kw_only=True)
class DoubleTriangular(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=([0, 0], [np.inf, 1])
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        def kernel(t: float) -> float:
            if 0 <= t and t <= b * a:
                return (t / a) ** 2 / b
            elif b * a <= t and t <= a:
                return 1 - (1 - t / a) ** 2 / (1 - b)
            else:
                return 0  # nan

        if isinstance(t, np.ndarray):
            return np.array(list(map(kernel, t)))
        else:
            return kernel(t)

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        def kernel(t: float) -> float:
            if 0 <= t and t <= b * a:
                return 2 * t / a / (a * b)
            elif b * a <= t and t <= a:
                return 2 * (1 - t / a) / (a * (1 - b))
            else:
                return 0  # nan

        if isinstance(t, np.ndarray):
            return np.array(list(map(kernel, t)))
        else:
            return kernel(t)

    def prepare(
        self, x: np.ndarray, y: np.ndarray, func: str = "cdf"
    ) -> tuple[np.ndarray, np.ndarray]:
        self.p0 = [x.max() + 10, 0.5]
        self.bounds = ([x.max(), 0], [np.inf, 1])
        return x, y



import dataclasses

import numpy as np
import numpy.typing as npt
import scipy.stats

from . import Model


@dataclasses.dataclass(init=True, kw_only=True)
class NormalGaussian(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=([-np.inf, 0], np.inf)
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        return scipy.stats.norm.cdf(x=t, loc=a, scale=b)

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        return scipy.stats.norm.pdf(x=t, loc=a, scale=b)



import dataclasses

import numpy as np
import numpy.typing as npt

from . import Model


@dataclasses.dataclass(kw_only=True)
class Polynomial(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=(-np.inf, [0, np.inf, np.inf])
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        return a * t**3 + b * t**2 + c * t

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        return 3 * a * t**2 + 2 * b * t + c



import os
import re
from pathlib import Path

import numpy as np
import pandas as pd


def read_csv(filepath: str | Path) -> tuple[int, np.ndarray, np.ndarray]:
    match_result = re.fullmatch(
        pattern=r"(?P<id>\d+).*", string=os.path.basename(filepath)
    )
    assert match_result
    id = int(match_result.group("id"))
    df = pd.read_csv(filepath)
    x = df["flowTime"].to_numpy()
    y = df["frequency"].to_numpy()
    return id, x, y


def prepare_y(x: np.ndarray, y: np.ndarray, func: str = "cdf") -> np.ndarray:
    if func == "cdf":
        return y.cumsum()
    elif func == "pdf":
        return y / (x[1] - x[0])
    else:
        raise NotImplementedError()



import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns


def plot_param_distribution(
    data: pd.DataFrame, x: str, fname: str, kind: str = "kde"
) -> None:
    plt.figure(dpi=600)
    sns.displot(data=data, x=x, kind=kind)
    plt.savefig(fname=fname)
    plt.close()


def plot_params_distribution(data: pd.DataFrame, fname: str, kind: str = "kde") -> None:
    params = filter(lambda column: len(column) == 1, data.columns)
    dfs: list[pd.DataFrame] = list()
    for param in params:
        df = pd.DataFrame()
        df["value"] = data[param]
        df["param"] = param
        dfs.append(df)
    df = pd.concat(dfs, ignore_index=True)
    plt.figure(dpi=600)
    sns.displot(
        data=df,
        x="value",
        col="param",
        kind=kind,
        facet_kws={"sharex": False, "sharey": False},
    )
    plt.savefig(fname=fname)
    plt.close()



import dataclasses

import numpy as np

from . import Model


@dataclasses.dataclass(kw_only=True)
class Rational(Model):
    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        return a * t / (1 + b * t + c * t**2)

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        return a * (1 - c * t**2) / (1 + b * t + c * t**2) ** 2



import dataclasses

import numpy as np

from . import Model


@dataclasses.dataclass(kw_only=True)
class Logistic(Model):
    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        s = a / (1 + b * np.exp(-c * t))
        return s

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        u = a * b * c * np.exp(-c * t) / (b + np.exp(c * t) ** 2)
        return u



import itertools
import os
import typing
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from rich.pretty import Pretty
from rich.table import Table

from ..curve_fitting import Model, curve_fitting
from ..curve_fitting.utils import prepare_y
from ..log import print


def prepare_plot(
    model: Model, x: np.ndarray, y: np.ndarray, fit: str = "cdf", pred: str = "cdf"
) -> tuple[float, pd.DataFrame]:
    res = curve_fitting(model=model, x=x, y=y, fits=[fit], preds=[pred])
    r2_score = typing.cast(dict, res["r2_scores"]).get((fit, pred), np.nan)
    popt = res["popts"][fit]
    pred_x = np.linspace(start=x.min(), stop=x.max(), num=1000)
    pred_y = model.select(pred)(pred_x, *popt)
    return r2_score, pd.concat(
        [
            pd.DataFrame(
                {
                    "x": x,
                    "y": prepare_y(x=x, y=y, func=pred),
                    "hue": "true",
                    "fit": fit,
                    "predict": pred,
                }
            ),
            pd.DataFrame(
                {
                    "x": pred_x,
                    "y": pred_y,
                    "hue": "predict",
                    "fit": fit,
                    "predict": pred,
                }
            ),
        ]
    )


def annotate_r2_score(
    data: pd.DataFrame, label: str, r2_scores: dict, *args, **kwargs
) -> None:
    if label == "true":
        return
    fit = data["fit"][0]
    pred = data["predict"][0]
    plt.text(
        x=data["x"].max(),
        y=(data["y"].min() + data["y"].max()) / 2,
        s=r"$R^2 = {:.5f}$".format(r2_scores[fit, pred]),
        horizontalalignment="right",
        usetex=True,
        verticalalignment="center",
    )


def plot_fitting(
    model: Model,
    x: np.ndarray,
    y: np.ndarray,
    output: str | Path,
    id: typing.Optional[int] = None,
) -> None:
    dfs: list[pd.DataFrame] = list()
    r2_scores: dict = dict()
    table = Table("fit", "predict", "r2_score", title=model.name)
    for fit, pred in itertools.product(["cdf", "pdf"], repeat=2):
        r2_score, df = prepare_plot(model=model, x=x, y=y, fit=fit, pred=pred)
        r2_scores[fit, pred] = r2_score
        table.add_row(fit, pred, Pretty(r2_score))
        dfs.append(df)
    print(table)
    data = pd.concat(dfs)
    plt.figure(dpi=300)
    g = sns.FacetGrid(
        data=data,
        row="predict",
        col="fit",
        hue="hue",
        sharex="col",
        sharey="row",
        legend_out=False,
    )
    g.map_dataframe(sns.lineplot, "x", "y")
    g.map_dataframe(annotate_r2_score, r2_scores=r2_scores)
    if id:
        g.add_legend(title=f"id = {id}")
    else:
        g.add_legend()
    g.tight_layout()
    os.makedirs(name=Path(output).parent, exist_ok=True)
    plt.savefig(output)
    plt.close()



import dataclasses

import numpy as np

from . import Model


@dataclasses.dataclass(kw_only=True)
class Multistage(Model):
    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        s = 1 - np.exp(-a * t - b * t**2 - c * t**3)
        return s

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        u = (a + 2 * b * t + 3 * c * t**2) * np.exp(-a * t - b * t**2 - c * t**3)
        return u



import dataclasses

import numpy as np
import numpy.typing as npt

from . import Model


@dataclasses.dataclass(kw_only=True)
class DoublePower(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=([0, 1, 1], np.inf)
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        s = (1 - (1 - t / a) ** b) ** c
        return s

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float, c: float) -> float | np.ndarray:
        u = (b * c / a) * (1 - t / a) ** (b - 1) * (1 - (1 - t / a) ** b) ** (c - 1)
        return u

    def prepare(
        self, x: np.ndarray, y: np.ndarray, func: str = "cdf"
    ) -> tuple[np.ndarray, np.ndarray]:
        self.p0 = [x.max() + 10, 2.0, 2.0]
        self.bounds = ([x.max(), 1.0, 1.0], np.inf)
        return x, y



from pathlib import Path

import click
import pandas as pd

from ..curve_fitting import NAME_TO_MODEL, Model
from ..curve_fitting.utils import read_csv
from ..log import print
from . import plot_fitting


def main(
    model: Model,
    param_path: str | Path,
    output: str | Path,
    key: str = "pdf_r2_score",
    threshold: float = 0.9,
    data_path: str | Path = "2-sub-WFIUH_rescaled",
):
    df = pd.read_csv(param_path)
    idx_max = df[key].idxmax()
    if df[key][idx_max] < threshold:
        print(f"{key} = {df[key][idx_max]} < {threshold}")
        return
    id = df["id"][idx_max]
    data_path = Path(data_path) / f"{id}_wfiuh.csv"
    id_, x, y = read_csv(data_path)
    assert id == id_
    plot_fitting(model=model, x=x, y=y, output=output, id=id)


@click.command("best-sample")
@click.option(
    "-m",
    "--model",
    type=click.Choice(choices=list(NAME_TO_MODEL.keys())),
    required=True,
)
@click.option(
    "--param-path",
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    required=True,
)
@click.option("-o", "--output", type=click.Path(), required=True)
@click.option("--key", default="pdf_r2_score")
@click.option("--threshold", type=click.FloatRange(max=1), default=0.9)
@click.option(
    "--data-path",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    default="2-sub-WFIUH_rescaled",
)
def command(
    model: str,
    param_path: str | Path,
    output: str | Path,
    key: str = "pdf_r2_score",
    threshold: float = 0.9,
    data_path: str | Path = "2-sub-WFIUH_rescaled",
):
    main(
        model=NAME_TO_MODEL[model](),
        param_path=param_path,
        output=output,
        key=key,
        threshold=threshold,
        data_path=data_path,
    )


if __name__ == "__main__":
    command()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wfiuh',
 'wfiuh.best_sample',
 'wfiuh.curve_fitting',
 'wfiuh.curve_fitting.models',
 'wfiuh.log',
 'wfiuh.param_distribution']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'matplotlib>=3.6.2,<4.0.0',
 'numpy>=1.23.4,<2.0.0',
 'pandas>=1.5.1,<2.0.0',
 'rich>=12.6.0,<13.0.0',
 'scikit-learn>=1.1.3,<2.0.0',
 'scipy>=1.9.3,<2.0.0',
 'seaborn>=0.12.1,<0.13.0']

setup_kwargs = {
    'name': 'wfiuh',
    'version': '0.3.0',
    'description': 'Curve fitting width function IUH (WFIUH) in Hydrology',
    'long_description': '# WFIUH\n\nCurve fitting width function IUH (WFIUH) in Hydrology\n\nResults can be found [here](https://drive.liblaf.top/share/WFIUH/results/).\n\n## CDF\n\n| Model              | Success Fits  |\n| ------------------ | ------------- |\n| Beta               | 11069 / 11069 |\n| DoublePower        | 11066 / 11069 |\n| DoubleTriangular   | 11069 / 11069 |\n| Frechet            | 11062 / 11069 |\n| Gamma              | 11062 / 11069 |\n| Hill               | 11059 / 11069 |\n| Hoerl              | 11069 / 11069 |\n| InverseGaussian    | 11066 / 11069 |\n| Kumaraswamy        | 11069 / 11069 |\n| Logistic           | 11062 / 11069 |\n| Multistage         | 11065 / 11069 |\n| NormalGaussian     | 11069 / 11069 |\n| Polynomial         | 11069 / 11069 |\n| Rational           | 11069 / 11069 |\n| ShiftedLogPearson3 | 11051 / 11069 |\n| Weibull            | 11069 / 11069 |\n\n## PDF\n\n| Model              | Success Fits  |\n| ------------------ | ------------- |\n| Beta               | 11061 / 11069 |\n| DoublePower        | 11065 / 11069 |\n| DoubleTriangular   | 11063 / 11069 |\n| Frechet            | 11039 / 11069 |\n| Gamma              | 11065 / 11069 |\n| Hill               | 11068 / 11069 |\n| Hoerl              | 11059 / 11069 |\n| InverseGaussian    | 11064 / 11069 |\n| Kumaraswamy        | 11069 / 11069 |\n| Logistic           | 10754 / 11069 |\n| Multistage         | 11065 / 11069 |\n| NormalGaussian     | 11069 / 11069 |\n| Polynomial         | 11069 / 11069 |\n| Rational           | 11064 / 11069 |\n| ShiftedLogPearson3 | 11027 / 11069 |\n| Weibull            | 11065 / 11069 |\n',
    'author': 'Qin Li',
    'author_email': 'liblaf@outlook.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://liblaf.github.io/WFIUH/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<3.12',
}


setup(**setup_kwargs)



import dataclasses

import numpy as np
import numpy.typing as npt
import scipy.stats

from . import Model

SCALE = 150


@dataclasses.dataclass(kw_only=True)
class Beta(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=([1, 0], np.inf)
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        s = scipy.stats.beta.cdf(t, a, b, scale=SCALE)
        return s

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        u = scipy.stats.beta.pdf(t, a, b, scale=SCALE)
        return u



import dataclasses

import numpy as np
import numpy.typing as npt

from . import Model

SCALE = 150


@dataclasses.dataclass(kw_only=True)
class Kumaraswamy(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=([0, 0], np.inf)
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        t /= SCALE
        s = 1 - (1 - t**a) ** b
        return s

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        t /= SCALE
        u = a * b * t ** (a - 1) * (1 - t**a) ** (b - 1)
        return u



import click

from .best_sample.__main__ import command as best_sample
from .curve_fitting.__main__ import command as curve_fitting
from .param_distribution.__main__ import main as param_distribution


@click.group(name="wfiuh", context_settings={"show_default": True})
def main() -> None:
    pass


main.add_command(cmd=best_sample)
main.add_command(cmd=curve_fitting)
main.add_command(cmd=param_distribution)


if __name__ == "__main__":
    main()



import dataclasses

import numpy as np
import numpy.typing as npt

from . import Model


@dataclasses.dataclass(kw_only=True)
class Frechet(Model):
    bounds: tuple[npt.ArrayLike, npt.ArrayLike] = dataclasses.field(
        default=([0, 0], np.inf)
    )

    @staticmethod
    def cdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        s = np.exp(-((t / b) ** -a))
        return s

    @staticmethod
    def pdf(t: float | np.ndarray, a: float, b: float) -> float | np.ndarray:
        u = a / b * (t / b) ** (-1 - a) * np.exp(-((t / b) ** (-a)))
        return u

    def prepare(
        self, x: np.ndarray, y: np.ndarray, func: str = "cdf"
    ) -> tuple[np.ndarray, np.ndarray]:
        return x[1:], y[1:]
